'''
Written by Sean West

Inputs:
    enrichment       enrichment type location = 'go', 'kegg', 'omim'
    input_loc        cluster file location
    output_loc       output file location
    all_nodes        file with all nodes in network
    update           which type of enrichment to update = 'go', 'kegg', 'omim'
'''
import getopt
import os

class Input_handler():
    def __init__(self, args):
        self.args = self.opts(args[1:])
        return
    
    def inputs(self):
        all_inputs = {}
        all_inputs['enrichment'] = self.args['e']
        all_inputs['input_loc'] = self.args['i']
        all_inputs['output_loc'] = self.args['o']
        all_inputs['all_nodes'] = self.args['n']
        all_inputs['update'] = self.args['u']
        return all_inputs
    
    def opts(self, options):
        shortops = 'e:i:o:n:u:'
        longops = ['enrichment=', 'input=', 'output=', 'nodes=', 'update=']
        opts = getopt.getopt(options, shortops, longops)
        
        args = {'e':'', 'i':'', 'o':'', 'n':'', 'u':''}
        for (opt, arg) in opts[0]:
            if opt == '-e' or opt == '--enrichment':
                args['e'] = arg
            elif opt == '-i' or opt == '--input':
                args['i'] = arg
            elif opt == '-o' or opt == '--output':
                args['o'] = arg
            elif opt == '-n' or opt == '--nodes':
                args['n'] = arg
            elif opt == '-u' or opt == '--update':
                args['u'] = arg
        return args




'''
Written by Sean West

SOURCE OBJECT
'''
import collections

class Source_ob():
    def __init__(self):
        self.groupings = collections.defaultdict(list)
        self.group_info = collections.defaultdict(dict)
        return


'''
Written by Sean West

User Defined Exceptions
'''


# update_enrichment
class UE_exception(Exception):
    def __init__(self):
        return
    
class NoOmimFile(UE_exception):
    def __init__(self):
        self.message = '\nOMIM does not let their files be automatically downloaded\n'
        self.message += 'Go to "www.omim.org/downloads" to request ftp access.\n'
        self.message += 'Then place these files in the cwd and rerun the update command:\n'
        self.message += '\tmim2gene.txt\n\tgenemap2.txt\n\n'


'''
Written by Sean West

Sets location for enrichment document in:
    enrichment_document
Parses:
    enrichment document       parse_ed
    input document            parse_id
    group info file           parse_gid
    node list file            parse_nodes
'''
import collections
import enrich


class Document_handler():
    def __init__(self):
        self.enrichment = ''
        self.e_doc = ''
        self.gi_doc = ''
        return
    
    def enrichment_document(self):
        path = (enrich.__file__).replace('__init__.py', '')
        eloc, dloc = ['','']
        if self.enrichment in ['GO', 'gene_ontology', 'geneontology', 'go']:
            eloc = path + 'docs/gene_ontology.groupings.txt'
            dloc = path + 'docs/gene_ontology.group_info.txt'
        elif self.enrichment in ['KEGG', 'kegg']:
            eloc = path +  'docs/kegg.groupings.txt'
            dloc = path + 'docs/kegg.group_info.txt'
        elif self.enrichment in ['OMIM', 'omim']:
            eloc = path + 'docs/omim.groupings.txt'
            dloc = path + 'docs/kegg.group_info.txt'
        return eloc, dloc
            
    def parse_ed(self):
        dic = collections.defaultdict(list)
        with open(self.e_doc) as ed:
            for line in ed:
                line = line.replace('\n', '')
                line = line.split('\t')
                first = line.pop(0)
                dic[first] = line
        return dic
    
    def parse_id(self, i_doc):
        clusters = collections.defaultdict(list)
        with open(i_doc) as id:
            i = 1
            for line in id:
                line = line.replace('\n', '')
                line = line.split('\t')
                clusters[i] = line
                i += 1
        return clusters
    
    def parse_gid(self):
        group_info = collections.defaultdict(dict)
        with open(self.gi_doc) as gd:
            for line in gd:
                line = line.replace('\n', '')
                line = line.split('\t')
                group_info[line[0]]['name'] = line[1]
        return group_info
    
    def parse_nodes(self, location):
        allnodes = []
        with open(location) as an:
            for line in an:
                allnodes.append(line.replace('\n', ''))
        return allnodes


'''
Written by Sean West

Uses hypergeometric distribution function to compare all the clusters to every group in source.
'''
import collections
from scipy.stats import hypergeom

class Stat_compare():
    def __init__(self):
        self.hits = collections.defaultdict(list)
        return
    
    def compare(self, clusters, groupings, allnodes):
        for cid in clusters:
            for gid in groupings:
                pval = self.hyper_geom_compare(clusters[cid], groupings[gid], allnodes)
                if pval <= 0.05:
                    self.hits[cid].append({'csize': len(clusters[cid]), 'gid':gid, 'pval':pval})
        return
                    
    def hyper_geom_compare(self, cluster, group, allnodes):
        common = len(set(cluster).intersection(group))
        g_in_net = len(set(group).intersection(allnodes))
        return hypergeom.sf(common, len(allnodes), g_in_net, len(cluster))


'''
Written by Sean West

This code is the core funtionality of the enrich program.

It:
        obtain inputs
                if 'update' then calls the update_enrichment
        creates enrichment dictionaries by parsing enrichment files:
                source.groupings.txt
                source.group_info.txt
        obtains input clusters
        compares clusters to source groups
        exports results
'''

from enrich import input_handler
from enrich import document_handler
from enrich import source_ob
from enrich import input_ob
from enrich import stat_compare
from enrich import export
from enrich import update_enrichment

import sys
from pkg_resources import resource_filename


class Central_hub():
    def __init__(self):
        return
    
    def main(self):
        # obtain inputs
        ih = input_handler.Input_handler(sys.argv)
        inputs = ih.inputs()
        enrichment = inputs['enrichment']
        input_loc = inputs['input_loc']
        outfile_loc = inputs['output_loc']
        all_nodes = inputs['all_nodes']
        
        # if need to update source files
        if inputs['update']:
            ue = update_enrichment.Update_enrichment()
            ue.update(inputs['update'])
            return
        
        # obtain enrichment dictionaries
        so = source_ob.Source_ob()
        dh = document_handler.Document_handler()
        
        dh.enrichment = enrichment
        dh.e_doc, dh.gi_doc = dh.enrichment_document()
        so.groupings = dh.parse_ed()
        so.group_info = dh.parse_gid()
        
        # obtain input clusters
        io = input_ob.Input_ob()
        io.nodes = dh.parse_nodes(all_nodes)
        io.clusters = dh.parse_id(input_loc)
        
        # stat compare
        sc = stat_compare.Stat_compare()
        sc.compare(groupings=so.groupings, clusters=io.clusters, allnodes=io.nodes)
        
        # export
        ex = export.Export()
        ex.report(sc.hits, so.group_info, outfile_loc)
        
        return
    
def smain():
    ''' Called by runner '''
    stick = Central_hub()
    stick.main()
    return



'''
Written by Sean West

Exports results as:
ClusterID    ClusterSize    GroupID    PVal    GroupName
'''

class Export():
    def __init__(self):
        return
    
    def report(self, hits, group_info, outfileloc):
        outfile = open(outfileloc, 'w')
        cid_order = self.most_hits(hits)
        for cid in cid_order:
            for outline in hits[cid]:
                gid = outline['gid']
                outfile.write('\t'.join([str(cid), str(outline['csize']), gid, str(outline['pval']), group_info[gid]['name']]) + '\n')
        outfile.close()        
        return
        
    def most_hits(self, hits):
        cid_numhits = []
        for cid in hits:
            cid_numhits.append((cid,len(hits[cid])))
        cid_numhits = sorted(cid_numhits, key=lambda x: x[1])
        return [x[0] for x in cid_numhits]


from setuptools import setup

files = ['input/*', 'out/*']

setup(
      name = 'west_enrich',
      version = '1',
      description = 'Local enrichment of cluster files (GO/OMIM)',
      author = 'Sean West',
      url = 'tbd',
      packages = ['enrich'],
      package_data = {'package' : files},
      entry_points = {
                      'console_scripts':[
                                         'west_enrich = enrich.central_hub:smain'
                                         ]
                      },
      scripts = ['run_enrich'],
      long_description = 'long description',
      classifiers = ['Programming Language :: Python', \
                     'Programming Language :: Python :: 3', \
                     'Operating System :: Unix', \
                     'Development Status :: 1 - Planning', \
                     'Intended Audience :: Science/Research', \
                     'Topic :: Scientific/Engineering :: Bio-Informatics', \
                     'Topic :: Scientific/Engineering :: Mathematics', \
                     'License :: OSI Approved :: GNU General Public License v3 (GPLv3)', \
                     'Natural Language :: English']
      )


import enrich
import subprocess
import gzip
import collections
import sys
import enrich.all_exceptions as aex


class Update_enrichment():
    def __init__(self):
        return
    
    def update(self, source):
        path = (enrich.__file__).replace('__init__.py', '')
        subprocess.call(['mkdir', path+'docs'])
        
        if source in ['go', 'gene_ontology', 'geneontology', 'go']:
            self.gene_ontology(path)
        if source in ['KEGG', 'kegg']:
            self.kegg(path)
        if source in ['OMIM', 'omim']:
            self.omim(path)
        if source in ['all', 'a', 'ALL', 'All']:
            self.gene_ontology(path)
            self.kegg(path)
            self.omim(path)
        return
    
    def gene_ontology(self, path):
        # retrieve source file
        sys.stdout.write('Retrieving Gene Ontology source file...\n\n')
        subprocess.call(['wget', 'ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/gene2go.gz'])
        gzfile = gzip.open('gene2go.gz', 'rb')
        gf = gzfile.read()
        gzfile.close()
        txtfile = open('gene2go', 'wb')
        txtfile.write(gf)
        txtfile.close()
        subprocess.call(['rm', 'gene2go.gz'])
        sys.stdout.write('Retrieving Gene Ontology source file...success\n')
        
        # parsing the source file
        sys.stdout.write('Parsing source file...')
        groupings = collections.defaultdict(list)
        groupinfo = collections.defaultdict(dict)
        infile = open('gene2go', 'r')
        infile.readline()
        for line in infile:
            line = line.replace('\n', '')
            tax_id, GeneID, GO_ID, Evidence, Qualifier, GO_term, PubMed, Category = line.split('\t')
            if tax_id == '9606':
                groupings[GO_ID].append(GeneID)
                if GO_ID not in groupinfo:
                    groupinfo[GO_ID]['name'] = GO_term
                    groupinfo[GO_ID]['category'] = Category
                    groupinfo[GO_ID]['taxid'] = tax_id
        subprocess.call(['rm', 'gene2go'])  
        sys.stdout.write('success\n')  
                
        # export to ouput files        
        sys.stdout.write('Exporting data files...')
        groupings_file = open(path+'docs/gene_ontology.groupings.txt', 'w')
        for go_id in groupings:
            groupings_file.write(go_id + '\t' + '\t'.join(groupings[go_id]) + '\n')
        groupings_file.close()
        
        groupinfo_file = open(path+'docs/gene_ontology.group_info.txt', 'w')
        for go_id in groupinfo:
            groupinfo_file.write('\t'.join([go_id, groupinfo[go_id]['taxid'], groupinfo[go_id]['name'], groupinfo[go_id]['category']]) + '\n')
        groupinfo_file.close()
        sys.stdout.write('success\n\n')
        return
    
    def kegg(self, path):
        sys.stdout.write('\nWe dont have KEGG\n\n')
        return
    
    def omim(self, path):
        try:
            mim2gene = open('mim2gene.txt', 'r')
            genemap2 = open('genemap2.txt', 'r')
        except:
            nomim = aex.NoOmimFile()
            sys.stdout.write(nomim.message)
            return
        
        # creating OMIM to GeneID reference
        sys.stdout.write('\nCreating OMIM to GeneID reference...')
        omim2geneid = {}
        mim2gene.readline()
        for line in mim2gene:
            line = line.split('\t')
            omim2geneid[line[0]] = line[2]
        mim2gene.close()
        sys.stdout.write('success\n')
        
        # creating omim groups
        sys.stdout.write('Parsing genemap2...')
        groupings = collections.defaultdict(list)
        groupinfo = collections.defaultdict(dict)
        for line in genemap2:
            line = line.replace('\n', '')
            line = line.split('|')
            gomim = line[8]
            disease_string = line[11]
            geneid = omim2geneid[gomim]

            domims = []
            if disease_string and disease_string != ' ' and geneid:
                disease_list = disease_string.split(';')
                for disease in disease_list:
                    disease_omim = ((disease.split(',')[-1]).split('(')[0]).replace(' ', '')
                    if not disease_omim:
                        disease_omim = ((disease.split(',')[-2]).split('(')[0]).replace(' ', '')
                    try:
                        int(disease_omim)
                    except: 
                        break
                    if len(disease_omim) != 6:
                        break
                    domims.append(disease_omim)
                    if disease_omim not in groupinfo:
                        groupinfo[disease_omim]['name'] = disease.strip()
                for domim in domims:
                    if geneid not in groupings[domim] and geneid != '-':
                        groupings[domim].append(geneid)
        sys.stdout.write('success\n')

        # export files        
        sys.stdout.write('Exporting data files...')
        groupings_file = open(path+'docs/omim.groupings.txt', 'w')
        for go_id in groupings:
            groupings_file.write(go_id + '\t' + '\t'.join(groupings[go_id]) + '\n')
        groupings_file.close()
        
        groupinfo_file = open(path+'docs/omim.group_info.txt', 'w')
        for go_id in groupinfo:
            groupinfo_file.write('\t'.join([go_id, '9606', groupinfo[go_id]['name'], '-']) + '\n')
        groupinfo_file.close()
        sys.stdout.write('success\n\n')
        
        return
    
    

        


from enrich.central_hub import Central_hub
from enrich.document_handler import Document_handler
from enrich.input_handler import Input_handler
from enrich.source_ob import Source_ob
from enrich.input_ob import Input_ob
from enrich.stat_compare import Stat_compare
from enrich.export import Export
from enrich.all_exceptions import UE_exception
from enrich.all_exceptions import NoOmimFile